// (c) 2000 Richard Grimes, iDevResource.com Ltd
// you are free to use this code as long as your copyright message
// acknowledges iDevResource.com Ltd

#ifndef _BUCKET_H
#define _BUCKET_H

// this is a 'bucket' that holds all the debug messages for a process
class CBucket
{
public:
   // this is the data of each message
   typedef struct _BucketData
   {
      DWORD cbSize;  // this is the size of the whole struct
      SYSTEMTIME st; // this is the time the message was logged
      TCHAR str[1];  // this is the message
   } BucketData, *LPBUCKETDATA;

   // this is an iterator to obtain data from the bucket class
   class iterator
   {
   public:
      iterator() : m_pData(NULL) {}
      iterator(LPBUCKETDATA pData) : m_pData(pData) {}
      // get the next item
      operator++(int)
      {
         m_pData = reinterpret_cast<LPBUCKETDATA>(reinterpret_cast<LPBYTE>(m_pData) + m_pData->cbSize);
      }
      // comparison operator
      bool operator<(iterator& rhs)
      {
         return m_pData < rhs.m_pData;
      }
      // dereference operator
      BucketData& operator*()
      {
         return *m_pData;
      }
   private:
      LPBUCKETDATA m_pData;
   };

   // process ID for this data
   DWORD m_pid;

   // constructor: allocate a Bucket for the specified process, 
   // initially cPages pages in size
   CBucket(DWORD pid, DWORD cPages = 1) : m_pid(pid), m_maxStr(0)
   {
      if (cPages == 0)
         cPages = 1;
      SYSTEM_INFO si;
      GetSystemInfo(&si);
      m_dwSize = si.dwPageSize * cPages;
      m_data = static_cast<LPBYTE>(malloc(m_dwSize));
      m_pos = reinterpret_cast<LPBUCKETDATA>(m_data);
   }

   // destructor
   ~CBucket()
   {
      if (m_data)
         free(m_data);
   }

   // Add another record to the bucket. str is the message string
   void Add(LPTSTR str)
   {
      // determine the amount of the buffer we will use, check to see if
      // the data has come from a file
      if (IsSavedData(str))
      {
         // don't save the time in the string
         m_pos->cbSize = (strlen(str) + 1) * sizeof(TCHAR) 
            + sizeof(DWORD) + sizeof(SYSTEMTIME) - 25;
      }
      else
      {
         m_pos->cbSize = (lstrlen(str) + 1) * sizeof(TCHAR) 
            + sizeof(DWORD) + sizeof(SYSTEMTIME);
      }

      // check to see if we have enough data in the bucket
      if ((reinterpret_cast<LPBYTE>(m_pos) - m_data 
           + m_pos->cbSize + sizeof(DWORD)) > m_dwSize)
      {
         // we need to allocate another page
         DWORD pos = reinterpret_cast<LPBYTE>(m_pos) - m_data;
         SYSTEM_INFO si;
         GetSystemInfo(&si);
         m_dwSize += si.dwPageSize;
         m_data = static_cast<LPBYTE>(realloc(m_data, m_dwSize));
         m_pos = reinterpret_cast<LPBUCKETDATA>(static_cast<LPBYTE>(m_data) + pos);
      }

      // store the time
      if (IsSavedData(str))
      {
         // get the time from the string
         str[4]  = 0; m_pos->st.wYear = _ttoi(str);
         str[7]  = 0; m_pos->st.wMonth = _ttoi(str + 5);
         str[10] = 0; m_pos->st.wDay = _ttoi(str + 8);
         str[13] = 0; m_pos->st.wHour = _ttoi(str + 11);
         str[16] = 0; m_pos->st.wMinute = _ttoi(str + 14);
         str[19] = 0; m_pos->st.wSecond = _ttoi(str + 17);
         str[23] = 0; m_pos->st.wMilliseconds = _ttoi(str + 20);
         // point to the actual string
         str += 25;
      }
      else
         GetLocalTime(&m_pos->st);

      // store the string
      lstrcpy(m_pos->str, str);
      if (m_pos->cbSize > m_maxStr)
         m_maxStr = m_pos->cbSize;
      // point to the insertion point of the next record
      m_pos = reinterpret_cast<LPBUCKETDATA>(reinterpret_cast<LPBYTE>(m_pos) + m_pos->cbSize);
      m_pos->cbSize = 0;
   }

   // returns the required record
   LPBUCKETDATA operator[](int iIndex)
   {
      LPBUCKETDATA pos = reinterpret_cast<LPBUCKETDATA>(m_data);
      while (iIndex > 0)
      {
         if (pos->cbSize == 0)
            return NULL;
         pos = reinterpret_cast<LPBUCKETDATA>(reinterpret_cast<LPBYTE>(pos) + pos->cbSize);
      }
      return pos;
   }

   // returns the first record in the buffer
   iterator begin() const
   {
      return iterator(reinterpret_cast<LPBUCKETDATA>(m_data));
   }

   // points to after the last record
   iterator end() const
   {
      return iterator(m_pos);
   }

   // Size of the largest record
   DWORD GetMax() const
   {
      return m_maxStr;
   }

   // this determines if the data was saved in a file and hence has a date
   static bool IsSavedData(LPTSTR str)
   {
      if (lstrlen(str) == 0) return false;
      return (str[23] == _T('$'));
   }
private:
   LPBYTE m_data;       // start of the buffer
   DWORD m_dwSize;      // size of the buffer (integral number of pages)
   LPBUCKETDATA m_pos;  // current position in the bucket
   DWORD m_maxStr;      // size of the largest record
};

#endif //_BUCKET_H
